/*
** Copyright 2017-2018 Double Precision, Inc.
** See COPYING for distribution information.
*/
#ifndef popup_handler_h
#define popup_handler_h

#include "popup/popup.H"
#include "generic_window_handler.H"
#include "x/w/connection_threadfwd.H"
#include "x/w/impl/grabbed_pointerfwd.H"
#include "x/w/rectangle.H"

#include <x/weakptr.H>

LIBCXXW_NAMESPACE_START

//! Return value from recalculate_popup_position

//! Indicates where recalculate_popup_position() placed the popup in relation
//! to its attached display element.
//!
//! \see popup

enum class popup_position_affinity {
	above,		//!< Popup positioned above its display element
	below,		//!< Popup positioned below its display element
	left,		//!< Popup positioned left of its display element
	right		//!< Popup positioned right of its display element
};

//! Popup handler implementation object.

//! Subclass generic_window's handler. Implement metrics methods to
//! invoke recalculate_popup_position(), then proceeds and sets the
//! top level display element's position accordingly.
//!
//! \see popup

class LIBCXX_HIDDEN popupObj::handlerObj :
	public generic_windowObj::handlerObj {

	//! Alias
	typedef generic_windowObj::handlerObj superclass_t;

public:

	//! Constructor
	handlerObj(ONLY IN_THREAD,
		   const ref<generic_windowObj::handlerObj> &parent,
		   const color_arg &background_color,
		   size_t nesting_level);

	//! Destructor
	~handlerObj();

	//! Override installed().
	void installed(ONLY IN_THREAD) override;

	//! Implement get_main_window()

	main_windowptr get_main_window() override;

	//! Override frame_extends_updated()

	//! This means we need to set_popup_position().

	void frame_extents_updated(ONLY IN_THREAD) override;

	//! Override horizvert_updated()

	//! This means we need to set_popup_position().

	void horizvert_updated(ONLY IN_THREAD) override;

	//! Implement theme_updated()
	void theme_updated(ONLY IN_THREAD, const defaulttheme &new_theme)
		override;

	//! Recalculate what my position should be.

	//! r is set to the default popup position, which the subclass can
	//! inspect and make the necessary adjustments.
	//!
	//! The end result gets adjusted to make sure it fits on the screen,
	//! then sets the popup position.

	virtual popup_position_affinity
		recalculate_popup_position(ONLY IN_THREAD,
					   rectangle &r,
					   dim_t screen_width,
					   dim_t screen_height)=0;

 protected:

	//! Call recalculate_popup_position(), then xcb_configure_window().

	//! Assemble all the information that recalculate_popup_position()
	//! needs to base its logic on, call it, then use the adjusted
	//! rectangle to xcb_configure_window().

	void set_popup_position(ONLY IN_THREAD);

 public:

	//! Override process_configure_notify()

	//! We manage our own size, and ignore ConfigureNotify messages
	//! from the server.
	void process_configure_notify(ONLY IN_THREAD,
				      const rectangle &) override;

	//! Override configure_notify_received()

	//! We manage our own size, and ignore ConfigureNotify messages
	//! from the server.
	void configure_notify_received(ONLY IN_THREAD,
				      const rectangle &) override;

	//! Override button_press_event

	//! Pressing a button outside of the popup hides it.
	void do_button_event(ONLY IN_THREAD,
			     const xcb_button_release_event_t *event,
			     const button_event &be,
			     const motion_event &me) override;

	//! Perform any needed grab/pointer processing due to visibility change.
	void set_inherited_visibility_mapped(ONLY IN_THREAD) override;

	//! Perform any needed grab/pointer processing due to visibility change.
	void set_inherited_visibility_unmapped(ONLY IN_THREAD) override;

	//! Override request_visibility()

	//! We resize/reposition ourselves only when we're visible.
	//! This optimizes recalculations that might occur several times
	//! when the popup is being created.
	//!
	//! Calls update_my_configuration().
	void request_visibility(ONLY IN_THREAD, bool flag) override;

	//! Override absolute_location_updated

	void absolute_location_updated(ONLY IN_THREAD,
				       absolute_location_update_reason reason)
		override;

 private:

	//! Our intended size.
	rectangle most_recent_configuration;

	//! Make sure most_recent_configuration is our actual configuration.

	//! Only when we're visible.
	void update_my_configuration(ONLY IN_THREAD);

	//! The pointer is grabbed while the popup is open.
	ptr<obj> current_grab;

	//! The parent window, which is used for the pointer grab.

	weakptr<ptr<generic_windowObj::handlerObj>> popup_parent;

	//! Override default_wm_class_resource().

	//! Look to the popup parent's.

	std::string default_wm_class_resource(ONLY IN_THREAD) override;

	//! Return the popup's parent main window.

	ptr<generic_windowObj::handlerObj> get_popup_parent(ONLY IN_THREAD)
		override;

	//! Mcguffin for a visible popup.

	//! This is returned by get_opened_mcguffin().

	ptr<obj> opened_mcguffin;

	//! Obtain a mcguffin for an opened, visible popup.

	//! The subclass must implement it by invoking the appropriate
	//! method in shared_handler_data, and returning the mcguffin.

	virtual ref<obj> get_opened_mcguffin(ONLY IN_THREAD)=0;

	//! We just released the opened_mcguffin

	//! Because we're no longer visible.
	//! The subclass must implement it by invoking the appropriate
	//! method in shared_handler_data.
	virtual void released_opened_mcguffin(ONLY IN_THREAD)=0;

 public:

	//! Perform processing after a popup is opened.

	//! Grab the pointer.
	virtual void popup_opened(ONLY IN_THREAD);

	//! The popup is about to be closed.

	//! Release the pointer grab.

	virtual void closing_popup(ONLY IN_THREAD);

	//! Override keep_passive_grab().

	//! Forward to parent element, that grabbed the pointer for us.

	bool keep_passive_grab(ONLY IN_THREAD) override;

	//! Override ungrab().

	//! Forward to parent element, that grabbed the pointer for us.
	void ungrab(ONLY IN_THREAD) override;

	//! Override grab_pointer().

	//! Forwarded to parent element.
	grabbed_pointerptr grab_pointer(ONLY IN_THREAD,
					const element_implptr &) override;

	//! Override process_key_event()

	//! If the ESC key is pressed, close the combo-box popup.

	bool process_key_event(ONLY IN_THREAD, const key_event &ke)
		override;
};

LIBCXXW_NAMESPACE_END

#endif
